﻿Imports System.Runtime.CompilerServices
Imports System.Text
Imports System.ComponentModel
Imports System.Text.RegularExpressions
Imports System.IO

Namespace System.Data

   Public Module DataTableX

#Region "SuspendTableEvents"

      Private _suspendTableEvents As New ComponentProperty(Of DataTable, ActionFrame)(AddressOf CreateActionFrame)

      Private Function CreateActionFrame(ByVal tb As DataTable) As ActionFrame
         Return New ActionFrame(AddressOf tb.BeginLoadData, AddressOf tb.EndLoadData)
      End Function

      <Extension()> _
      Public Function EnterBeginLoadData(ByVal tb As DataTable) As IDisposable
         Return _suspendTableEvents(tb).Enter()
      End Function

#End Region 'SuspendTableEvents

#Region "Joiner"
      'ein Sql-Identitäts-InnerJoin verknüpft die Ergebnismengen zweier Abfragen. Von beiden Abfragen wird je eine Spalte bestimmt, und diejenigen Datensätze werden in die Verknüpfung selektiert, bei denen die Spalten-Werte beider Abfragen übereinstimmen (Identität).
      'Eine InnerJoin-Abfrage kann selbst wieder mit einer weiteren Abfrage gejoint werden, usw.
      'In Sql können Spaltenwerte aller abgefragten Tabellen in die Ergebnismenge hineingenommen werden, im Dataset ist dies nicht sinnvoll, da solche "Aus 2 mach 1"-Tabellen nicht abgespeichert werden können (Im Dataset lädt man lieber die Daten in getrennte Tabellen, die dann ggfs. verknüpft angezeigt werden können, aber dennoch als Tabellen getrennt speicherbar sind).
      'Joiner generiert den From-Abschnitt solcher InnerJoins.
      'Die Zusatzbedingungen werden als Strings formuliert, Spaltenbezeichner können sich nur auf die Tabelle der DataColumn beziehen, die der Expression vorangeht. Namentliches Qualifizieren der Herkunftstabelle führt zu Fehlern.
      'Bsp: gegeben die Tabellen Person, Foto, FotoStichwort
      'Person.PersonIDColumn.JoinWith("PersonID=?", Foto.PersonIDColumn).AddJoin(Foto.FotoIDColumn, FotoStichwort.FotoIDColumn).Sql generiert dieses SQL-Fragment:
      '"((Person INNER JOIN Foto ON (Person.PersonID =? AND Person.PersonID = Foto.PersonID )) INNER JOIN FotoStichwort ON (Foto.FotoID = FotStichwort.FotoID))"
      'Dem kann man "SELECT FotoStichwort.* FROM " voranstellen, und erhält alle FotoStichworte aller Fotos der Person mit PersonID =? (parametrisierter Command).
      'V.a. aber sind die Spaltenbezeichner bei im Join eingearbeiteten Expressions den Tabellen beim Generieren eindeutig zuordnebar, auch wenn Tabellen umbenannt wurden.
      'Eine angehängte WHERE - Klausel hingegen müsste vollqualifizierte SpaltenNamen verwenden, da die Umbenennung aber mit-generiert wird, müsste der den Joiner nutzende Progger die generierten TableNamen vorhersehen - geht schier gar net.

      Public Interface IJoiner
         ''' <summary> Spaltenangaben in expression1 beziehen sich auf die Tabelle von clmn1 </summary>
         Function AddJoin(ByVal clmn0 As DataColumn, ByVal clmn1 As DataColumn, Optional ByVal expression1 As String = Nothing) As IJoiner
         ''' <summary>
         '''  Spaltenangaben in expression0 beziehen sich auf die Tabelle von clmn0.
         '''  Spaltenangaben in expression1 beziehen sich auf die Tabelle von clmn1. </summary>
         Function AddJoin(ByVal clmn0 As DataColumn, ByVal expression0 As String, ByVal clmn1 As DataColumn, Optional ByVal expression1 As String = Nothing) As IJoiner
         Function Sql() As String
         Sub Clear()
      End Interface

      Public Class Joiner : Inherits List(Of [Set](Of DataColumn, String)()) : Implements IJoiner

         ''' <summary>
         '''  Spaltenangaben in expression1 beziehen sich auf die Tabelle von clmn1. </summary>
         Public Function AddJoin(ByVal clmn0 As DataColumn, ByVal clmn1 As DataColumn, Optional ByVal expression1 As String = Nothing) As IJoiner Implements IJoiner.AddJoin
            Return AddJoin(clmn0, Nothing, clmn1, expression1)
         End Function

         ''' <summary>
         '''  Spaltenangaben in expression0 beziehen sich auf die Tabelle von clmn0.
         '''  Spaltenangaben in expression1 beziehen sich auf die Tabelle von clmn1. </summary>
         Public Function AddJoin(ByVal clmn0 As DataColumn, ByVal expression0 As String, ByVal clmn1 As DataColumn, Optional ByVal expression1 As String = Nothing) As IJoiner Implements IJoiner.AddJoin
            Add(NewArray(NewSet(clmn0, expression0), NewSet(clmn1, expression1)))
            Return Me
         End Function

         Private Sub Clear1() Implements IJoiner.Clear
            MyBase.Clear()
         End Sub

         Public Function Sql() As String Implements IJoiner.Sql
            Dim knownClmns As IList(Of String) = New List(Of String)
            Dim counter = 0
            Dim joinInfo As [Set](Of DataColumn, String)()
            Dim tbls(1) As String
            tbls(0) = Me(0)(0).A.Table.TableName
            Dim sb = New StringBuilder(New String("("c, Me.Count))
            sb.Add(tbls(0))
            knownClmns.Add(tbls(0))
            For Each joinInfo In Me
               'ein joinInfo besteht aus 2 [Set](Of DataColumn, String), z.B. eine ParentColumn und eine expression für die ParentTable, und eine ChildColumn und expression für die ChildTable.
               Dim clmn = Function(n As Integer) tbls(n).And(".", joinInfo(n).A.ColumnName)
               tbls(1) = joinInfo(1).A.Table.TableName
               Dim renaming = tbls(1)
               If knownClmns.Contains(renaming) Then          'ggfs. renaming einfügen
                  tbls(1) = "X" & counter
                  counter += 1
                  renaming &= " AS " & tbls(1)
               Else
                  knownClmns.Add(renaming)
               End If
               sb.Add(" INNER JOIN ", renaming, " ON (")
               For i = 0 To 1
                  'ggfs. ParentTable-/ChildTable-expressions dranfrickeln
                  If joinInfo(i).B IsNot Nothing Then sb.Add(tbls(i), ".", joinInfo(i).B, " AND ")
               Next
               sb.Add(clmn(0), "=", clmn(1), "))") 'Columns joinen
               tbls(0) = tbls(1)
            Next
            Return sb.ToString
         End Function

      End Class 'JoinWith

      ''' <summary>
      '''  Spaltenangaben in expression1 beziehen sich auf die Tabelle von clmn1. </summary>
      <Extension()> _
      Public Function JoinWith(ByVal clmn0 As DataColumn, ByVal clmn1 As DataColumn, Optional ByVal expression1 As String = Nothing) As IJoiner
         Return (New Joiner).AddJoin(clmn0, clmn1, expression1)
      End Function

      ''' <summary>
      '''  Spaltenangaben in expression0 beziehen sich auf die Tabelle des JoinWith()-Subjekts.
      '''  Spaltenangaben in expression1 beziehen sich auf die Tabelle von clmn1. </summary>
      <Extension()> _
      Public Function JoinWith(ByVal clmn0 As DataColumn, ByVal expression0 As String, ByVal clmn1 As DataColumn, Optional ByVal expression1 As String = Nothing) As IJoiner
         Return (New Joiner).AddJoin(clmn0, expression0, clmn1, expression1)
      End Function

#End Region 'Joiner


      <Extension()> _
      Public Function All(Of T As DataRow)(ByVal subj As TypedTableBase(Of T)) As IEnumerable(Of T)
         Return DirectCast(subj, DataTable).All.Cast(Of T)()
      End Function
      <Extension()> _
      Public Function All(ByVal subj As DataTable) As IEnumerable(Of DataRow)
         Return From rw In subj Where 0 = (rw.RowState And (DataRowState.Deleted Or DataRowState.Detached))
      End Function

      <Extension()> _
      Public Function ChildRels(ByVal subj As DataTable) As IEnumerable(Of DataRelation)
         Return subj.ChildRelations.Cast(Of DataRelation)()
      End Function

      <Extension()> _
      Public Function ParentRels(ByVal subj As DataTable) As IEnumerable(Of DataRelation)
         Return subj.ParentRelations.Cast(Of DataRelation)()
      End Function

      <Extension()> _
      Public Function ChildTables(ByVal subj As DataTable) As IEnumerable(Of DataTable)
         Return From rl In subj.ChildRelations.Cast(Of DataRelation)() _
            Where rl.ChildKeyConstraint.NotNull Select rl.ChildTable
         Return subj.ChildRelations.Cast(Of DataRelation).Select(Function(rl) rl.ChildTable)
      End Function

      <Extension()> _
      Public Function ParentTables(ByVal subj As DataTable) As IEnumerable(Of DataTable)
         Return subj.ParentRelations.Cast(Of DataRelation).Select(Function(rl) rl.ParentTable)
      End Function

      <Extension()> _
      Public Sub ClearRecursive(ByVal subj As DataTable)
         subj.ChildTables.ForEach(AddressOf ClearRecursive)
         subj.Clear()
      End Sub

      ''' <summary>
      ''' als Filler die parameterlose Fill-Methode eines generierten TableAdapters angeben
      ''' </summary>
      <Extension()> _
      Public Sub FillX(Of T As DataTable)(ByVal subj As T, ByVal Filler As Action(Of T))
         Using subj.EnterBeginLoadData
            Filler(subj)
         End Using
      End Sub

      ''' <summary>
      ''' als Filler die FillBy-Methode eines generierten TableAdapters angeben, die untergeordnete Datensätze abruft
      ''' </summary>
      <Extension()> _
      Public Sub FillBy(Of T As DataTable, Tid)(ByVal subj As T, ByVal Filler As Action(Of T, Tid), ByVal ID As Tid)
         Using subj.EnterBeginLoadData
            Filler(subj, ID)
         End Using
      End Sub

      ''' <summary>eases to implement standard-reactions on Row -init, -change, -delete</summary>
      <Extension()> _
      Public Sub WireUp(Of T As DataRow)(ByVal table As TypedTableBase(Of T), Optional ByVal onNewRow As Action(Of T) = Nothing, Optional ByVal onChange As Action(Of T) = Nothing, Optional ByVal onDelete As Action(Of T) = Nothing)
         If onNewRow.NotNull() Then
            AddHandler table.TableNewRow, Sub(s, e) onNewRow(DirectCast(e.Row, T))
         End If
         If onChange.NotNull() Then
            AddHandler table.RowChanged, _
               Sub(s, e)
                  If 0 < (e.Action And (DataRowAction.Add Or DataRowAction.Change)) Then
                     onChange(DirectCast(e.Row, T))
                  End If
               End Sub
         End If
         If onDelete.NotNull() Then
            AddHandler table.RowDeleting, Sub(s, e) onDelete(DirectCast(e.Row, T))
         End If
      End Sub

   End Module

End Namespace